import concurrent.futures
import logging
import time
from datetime import timedelta

from concurrentevents import _base
from concurrentevents._exceptions import Cancel
from concurrentevents.event import Event, Start, Exit


class EventManager:
    """
    A bare bones event framework using concurrent futures for threading

    Kwargs:
        :param threads: The number of threads, default to none for automatic calculation
        :type threads: int, optional
        :param verbosity: The level of logs expected from the event system
        :type verbosity: class:`concurrentevents.Verbosity`, int, optional
        :param logger: A logging object for customized logging
        :type logger: logging.RootLogger, optional
        :param process_tracking: Boolean for output of event system progress to logger
        :type process_tracking: bool, optional
        :param debug: Boolean for larger debug in the main thread
        :type debug: bool, optional
    """
    def __init__(self,
                 threads=None,
                 logger=None,
                 process_tracking=False,
                 debug=False):
        """Constructor Method"""
        _base._log = logger or logging.getLogger("__main__")

        # Futures Dictionary
        # This is a private dictionary used internally so that the main thread can manage and clean
        #   the running futures generated by the event system
        # Outline:
        #   key: Future Object
        #   value: (fn, args, kwargs)
        self.__futures = {}

        self.__thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=threads)

        if threads:
            _base._threads = threads
        else:
            _base._threads = self.__thread_pool._max_workers

        # This boolean is set so that after firing the start event on the main thread there
        #   are no other instances of it that could be created and thus no duplicate cleaners
        self.started = False

        # This boolean is set so that any dedicated while loops not made by the Thread Pool Executor
        #   will be able to check if the program is meant to end
        self.exit = False

        # Option to enable logging of progress tracking for all events needed to run
        self.progress_tracking = process_tracking
        if self.progress_tracking:
            self.fired = 0
            self.finished = 0
            self.errors = 0
            self.canceled = 0

        # Debug boolean for main loop debug statements
        self.debug = debug

        # Set a reference to the instantiated eventManager so other parts of this library can access it
        #   without it being passed around
        _base._event_manager = self

    def __handle(self, event):
        """
        Operates function specified to handle a specific event on that event

        This is the basis for the entire event system in terms of handling events in order
        and making sure the exit event ends the program after it runs

        Args:
            :param event: An event object to be handled
            :type event: class:`concurrentevents.Event`
        """
        eventName = event.__class__.__name__

        if isinstance(event, Exit):
            self.exit = event.timeout

        if eventName in _base._handlers.keys():
            handlers = sorted(_base._handlers[eventName], key=lambda h: h.priority)
            _base._log.info(f" {[h.__name__ for h in handlers]} handling {event.__str__()}")

            for h in handlers:
                try:
                    h(*event.args, **event.kwargs)
                except Cancel:
                    if self.progress_tracking:
                        self.canceled += 1
                    break

    def _submit(self, f, *args, **kwargs):
        """
        Interacts with the threadpool and futures dictionary to organize and track events

        This is implemented as the only methodology to put a function into the threadpool because,
        it forces all futures generated by putting something in the threadpool to be added to the futures list

        Parameters:
            :param f: A function object
            :type f: class:`function`
            :param `*args`: Arguments for f
            :param `**kwargs`: Keyword arguments for f
        """
        future = self.__thread_pool.submit(f, *args, **kwargs)
        self.__futures[future] = (f, args, kwargs)
        return future

    def fire(self, event):
        """
        A filter method for events that checks to see if it needs to be sent to the Thread Pool

        The purpose of this function is to validate not modify anything that wanted to be fired as an event.
        This intermediary step helps to eliminate anything that gets fired but isn't an event and any events that
        don't have handlers

        Args:
            :param event: An event that was fired
            :type event: Event


        :raises ValueError: Raised for when a non event is passed into the event param
        :raises KeyError: Raised for when there no handlers for the event
        """
        if not isinstance(event, Event):
            raise ValueError

        # Check to see if there are handlers ready to catch the event
        if event.__class__.__name__ not in _base._handlers.keys() and not isinstance(event, Exit):
            raise KeyError(f"No Handlers For {event.__class__.__name__}")

        # DEBUG
        if not self.exit:
            _base._log.debug(f"Firing {event}")

            self._submit(self.__handle, event)
            if self.progress_tracking:
                self.fired += 1

    def start(self):
        """
        Starts the event sequence

        Fires the start event to trigger the beginning of all functionality
        After firing the start event the main while loop is started for debug and cleaning of completed futures
        """
        if self.started:
            return

        s = time.time()

        self.fire(Start())

        while not (self.exit or len(self.__futures.keys()) == 0):
            try:
                if self.debug:
                    process_info = f"EventManager Loop (futures={len(list(self.__futures))}, " \
                                   f"running={len([ft for ft in list(self.__futures.keys()) if ft.running()])}, " \
                                   f"monitored={_base._thread_monitor})"
                    _base._log.info(process_info)

                for f, fn in self.__futures.copy().items():
                    if f.done():
                        self.__futures.pop(f)

                        try:
                            if self.progress_tracking:
                                self.finished += 1

                                future_info = fn[1][0]
                                if fn[0].__name__ != '__handle':
                                    future_info = f"{fn[0].__name__} (args={fn[1]}, kwargs={fn[2]})"
                                _base._log.debug(f"{future_info} ({self.finished}/{self.fired})")

                            # Raises errors
                            f.result()
                        except Exception:
                            _base._log.exception(fn)
                            if self.progress_tracking:
                                self.errors += 1

                time.sleep(0.25)
            except Exception:
                _base._log.exception("Event Manager Main Loop Error")

        if not self.exit:
            self.fire(Exit())
            print("Firing Exit")

        try:
            concurrent.futures.wait(self.__futures.keys(), timeout=self.exit if self.exit else 300)
            if self.progress_tracking:
                self.finished += len(self.__futures.keys())
        except TimeoutError:
            _base._log.exception("Final Cleaning Timed Out")

        _base._log.debug(f"Runtime {timedelta(seconds=time.time() - s)}")
        if self.progress_tracking:
            _base._log.debug(f"Events Fired ({self.fired})")

            _base._log.debug(f"Events Finished ({self.finished})")
            _base._log.debug(f"Events Errors ({self.errors})")

            _base._log.debug(f"Events Canceled ({self.canceled})")

        self.__thread_pool.shutdown()
